@@ -196,12 +196,14 @@ int mroute4_enable(int do_vifs, int table_id, int timeout)
 	struct iface *iface;
 	static int running = 0;
 
-	mroute4_socket = socket_create(AF_INET, SOCK_RAW, IPPROTO_IGMP, handle_nocache4, NULL);
 	if (mroute4_socket < 0) {
-		if (ENOPROTOOPT == errno)
-			smclog(LOG_WARNING, "Kernel does not even support IGMP, skipping ...");
+		mroute4_socket = socket_create(AF_INET, SOCK_RAW, IPPROTO_IGMP, handle_nocache4, NULL);
+		if (mroute4_socket < 0) {
+			if (ENOPROTOOPT == errno)
+				smclog(LOG_WARNING, "Kernel does not even support IGMP, skipping ...");
 
-		return -1;
+			return -1;
+		}
 	}
 
 #ifdef MRT_TABLE /* Currently only available on Linux  */
@@ -238,7 +240,7 @@ int mroute4_enable(int do_vifs, int table_id, int timeout)
 	/* Initialize virtual interface table */
 	memset(&vif_list, 0, sizeof(vif_list));
 
-	/* Create virtual interfaces (VIFs) for all non-loopback interfaces supporting multicast */
+	/* Create virtual interfaces (VIFs) for all IFF_MULTICAST interfaces */
 	if (do_vifs) {
 		for (iface = iface_iterator(1); iface; iface = iface_iterator(0))
 			mroute4_add_vif(iface);
@@ -267,7 +269,7 @@ error:
  *
  * Disable IPv4 multicast routing and release kernel routing socket.
  */
-void mroute4_disable(void)
+void mroute4_disable(int close_socket)
 {
 	struct mroute4 *entry, *tmp;
 
@@ -278,8 +280,10 @@ void mroute4_disable(void)
 	if (setsockopt(mroute4_socket, IPPROTO_IP, MRT_DONE, NULL, 0))
 		smclog(LOG_WARNING, "Failed shutting down IPv4 multicast routing socket: %s", strerror(errno));
 
-	socket_close(mroute4_socket);
-	mroute4_socket = -1;
+	if (close_socket) {
+		socket_close(mroute4_socket);
+		mroute4_socket = -1;
+	}
 
 	/* Free list of (*,G) routes on SIGHUP */
 	LIST_FOREACH_SAFE(entry, &mroute4_conf_list, link, tmp) {
@@ -353,7 +357,10 @@ static int mroute4_add_vif(struct iface *iface)
 	iface->vif = vif;
 	vif_list[vif].iface = iface;
 
-	return mrdisc_register(iface->name, vif);
+	if (iface->mrdisc)
+		return mrdisc_register(iface->name, vif);
+
+	return 0;
 }
 
 static int mroute4_del_vif(struct iface *iface)
@@ -380,7 +387,10 @@ static int mroute4_del_vif(struct iface *iface)
 	else
 		iface->vif = -1;
 
-	return mrdisc_deregister(vif);
+	if (iface->mrdisc)
+		return mrdisc_deregister(vif);
+
+	return 0;
 }
 
 /* Actually set in kernel - called by mroute4_add() and mroute4_check_add() */
@@ -526,26 +536,33 @@ int mroute4_dyn_add(struct mroute4 *route)
 		memset(route->ttl, 0, NELEMS(route->ttl) * sizeof(route->ttl[0]));
 	}
 
-	/* Add to list of dynamically added routes. Necessary if the user
+	ret = kern_add4(route, entry ? 1 : 0);
+	if (ret)
+		return ret;
+
+	/*
+	 * Add to list of dynamically added routes. Necessary if the user
 	 * removes the (*,G) using the command line interface rather than
 	 * updating the conf file and SIGHUP. Note: if we fail to alloc()
-	 * memory we don't do anything, just add kernel route silently. */
+	 * memory we don't do anything, just add kernel route silently.
+	 */
 	new_entry = malloc(sizeof(struct mroute4));
 	if (new_entry) {
 		memcpy(new_entry, route, sizeof(struct mroute4));
 		LIST_INSERT_HEAD(&mroute4_dyn_list, new_entry, link);
 	}
 
-	ret = kern_add4(route, entry ? 1 : 0);
+	/* Signal to cache handler we've added a stop filter */
 	if (!entry) {
 		errno = ENOENT;
-		ret = -1;
+		return -1;
 	}
-	return ret;
+
+	return 0;
 }
 
-/* 
- * Query kernel for install MFC entry usage statistics
+/*
+ * Query kernel for route usage statistics
  */
 static int get_stats4(struct mroute4 *route, unsigned long *pktcnt, unsigned long *bytecnt, unsigned long *wrong_if)
 {
@@ -663,6 +680,21 @@ static int mroute4_exists(struct mroute4 *route)
 	return 0;
 }
 
+/* Only inbound differs, there can only be one ... */
+static struct mroute4 *mroute4_similar(struct mroute4 *route)
+{
+	struct mroute4 *entry;
+
+	LIST_FOREACH(entry, &mroute4_static_list, link) {
+		if (entry->source.s_addr == route->source.s_addr &&
+		    entry->group.s_addr  == route->group.s_addr  &&
+		    entry->len           == route->len)
+			return entry;
+	}
+
+	return NULL;
+}
+
 /**
  * mroute4_add - Add route to kernel, or save a wildcard route for later use
  * @route: Pointer to struct mroute4 IPv4 multicast route to add
@@ -678,11 +710,20 @@ int mroute4_add(struct mroute4 *route)
 {
 	struct mroute4 *entry;
 
+	/* Exact match, then skip ... */
 	if (mroute4_exists(route)){
 		errno = EEXIST;
 		return 1;
 	}
 
+	/* ... (S,G) matches and inbound differs, then replace route */
+	entry = mroute4_similar(route);
+	if (entry) {
+		kern_del4(entry, is_active4(entry));
+		LIST_REMOVE(entry, link);
+		free(entry);
+	}
+
 	entry = malloc(sizeof(struct mroute4));
 	if (!entry) {
 		smclog(LOG_WARNING, "Cannot add multicast route: %s", strerror(errno));
@@ -703,10 +744,16 @@ int mroute4_add(struct mroute4 *route)
 		/* Also, immediately expire any currently blocked traffic */
 		LIST_FOREACH_SAFE(dyn, &mroute4_dyn_list, link, tmp) {
 			if (!is_active4(dyn) && is_match4(entry, dyn)) {
+				char origin[INET_ADDRSTRLEN], group[INET_ADDRSTRLEN];
+
+				inet_ntop(AF_INET, &dyn->group,  group,  INET_ADDRSTRLEN);
+				inet_ntop(AF_INET, &dyn->source, origin, INET_ADDRSTRLEN);
+				smclog(LOG_DEBUG, "Flushing (%s,%s) on VIF %d, new matching (*,G) rule ...",
+				       origin, group, dyn->inbound);
+
 				kern_del4(dyn, 0);
 				LIST_REMOVE(dyn, link);
 				free(dyn);
-				break;
 			}
 		}
 
@@ -717,6 +764,22 @@ int mroute4_add(struct mroute4 *route)
 	return kern_add4(route, 1);
 }
 
+/* Remove from kernel and linked list */
+static int do_mroute4_del(struct mroute4 *entry)
+{
+	int ret;
+
+	ret = kern_del4(entry, is_active4(entry));
+	if (ret && ENOENT != errno)
+		return ret;
+
+	/* Also remove on ENOENT */
+	LIST_REMOVE(entry, link);
+	free(entry);
+
+	return ret;
+}
+
 /**
  * mroute4_del - Remove route from kernel, or all matching routes if wildcard
  * @route: Pointer to struct mroute4 IPv4 multicast route to remove
@@ -737,33 +800,46 @@ int mroute4_del(struct mroute4 *route)
 			if (!is_exact_match4(entry, route))
 				continue;
 
-			LIST_REMOVE(entry, link);
-			free(entry);
-			break;
+			return do_mroute4_del(entry);
 		}
 
-		return kern_del4(route, 1);
+		/* Not found in static list, check if spawned from a (*,G) rule. */
+		LIST_FOREACH_SAFE(entry, &mroute4_dyn_list, link, tmp) {
+			if (!is_exact_match4(entry, route))
+				continue;
+
+			return do_mroute4_del(entry);
+		}
+
+		errno = ENOENT;
+		return -1;
 	}
 
 	/* Find matching (*,G) ... and interface .. and prefix length. */
 	LIST_FOREACH_SAFE(entry, &mroute4_conf_list, link, tmp) {
+		int ret = 0;
+
 		if (!is_match4(entry, route) || entry->len != route->len)
 			continue;
 
+		/* Remove all (S,G) routes spawned from the (*,G) as well ... */
 		LIST_FOREACH_SAFE(set, &mroute4_dyn_list, link, tmp) {
 			if (!is_match4(entry, set) || entry->len != route->len)
 				continue;
 
-			kern_del4(set, is_active4(set));
-			LIST_REMOVE(set, link);
-			free(set);
+			ret += do_mroute4_del(set);
 		}
 
-		LIST_REMOVE(entry, link);
-		free(entry);
+		if (!ret) {
+			LIST_REMOVE(entry, link);
+			free(entry);
+		}
+
+		return ret;
 	}
 
-	return 0;
+	errno = ENOENT;
+	return -1;
 }
 
 #ifdef HAVE_IPV6_MULTICAST_ROUTING
@@ -821,15 +897,16 @@ int mroute6_enable(int do_vifs, int table_id)
 	(void)table_id;
 #else
 	int arg = 1;
-	unsigned int i;
 	struct iface *iface;
 
-	mroute6_socket = socket_create(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6, handle_nocache6, NULL);
 	if (mroute6_socket < 0) {
-		if (ENOPROTOOPT == errno)
-			smclog(LOG_WARNING, "Kernel does not even support IPv6 ICMP, skipping ...");
+		mroute6_socket = socket_create(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6, handle_nocache6, NULL);
+		if (mroute6_socket < 0) {
+			if (ENOPROTOOPT == errno)
+				smclog(LOG_WARNING, "Kernel does not even support IPv6 ICMP, skipping ...");
 
-		return -1;
+			return -1;
+		}
 	}
 
 #ifdef MRT6_TABLE /* Currently only available on Linux  */
@@ -876,7 +953,7 @@ int mroute6_enable(int do_vifs, int table_id)
 		}
 	}
 #endif
-	/* Create virtual interfaces, IPv6 MIFs, for all non-loopback interfaces */
+	/* Create virtual interfaces, IPv6 MIFs, for all IFF_MULTICAST interfaces */
 	if (do_vifs) {
 		for (iface = iface_iterator(1); iface; iface = iface_iterator(0))
 			mroute6_add_mif(iface);
@@ -896,7 +973,7 @@ error:
  *
  * Disable IPv6 multicast routing and release kernel routing socket.
  */
-void mroute6_disable(void)
+void mroute6_disable(int close_socket)
 {
 #ifdef HAVE_IPV6_MULTICAST_ROUTING
 	if (mroute6_socket < 0)
@@ -905,8 +982,10 @@ void mroute6_disable(void)
 	if (setsockopt(mroute6_socket, IPPROTO_IPV6, MRT6_DONE, NULL, 0))
 		smclog(LOG_WARNING, "Failed shutting down IPv6 multicast routing socket: %s", strerror(errno));
 
-	socket_close(mroute6_socket);
-	mroute6_socket = -1;
+	if (close_socket) {
+		socket_close(mroute6_socket);
+		mroute6_socket = -1;
+	}
 #endif /* HAVE_IPV6_MULTICAST_ROUTING */
 }
 
@@ -1067,42 +1146,49 @@ int mroute6_del(struct mroute6 *route)
 #endif /* HAVE_IPV6_MULTICAST_ROUTING */
 
 /* Used by file parser to add VIFs/MIFs after setup */
-int mroute_add_vif(char *ifname, uint8_t threshold)
+int mroute_add_vif(char *ifname, uint8_t mrdisc, uint8_t threshold)
 {
-	int ret;
+	int ret = 0;
 	struct iface *iface;
+	struct ifmatch state;
 
 	smclog(LOG_DEBUG, "Adding %s to list of multicast routing interfaces", ifname);
-	iface = iface_find_by_name(ifname);
-	if (!iface)
-		return 1;
-
-	iface->threshold = threshold;
-	ret = mroute4_add_vif(iface);
+	iface_match_init(&state);
+	while ((iface = iface_match_by_name(ifname, &state))) {
+		iface->mrdisc    = mrdisc;
+		iface->threshold = threshold;
+		ret += mroute4_add_vif(iface);
 #ifdef HAVE_IPV6_MULTICAST_ROUTING
-	ret += mroute6_add_mif(iface);
+		ret += mroute6_add_mif(iface);
 #endif
+	}
 
-	return ret;
+	if (!state.match_count)
+		return 1;
+	else
+		return ret;
 }
 
 /* Used by file parser to remove VIFs/MIFs after setup */
 int mroute_del_vif(char *ifname)
 {
-	int ret;
+	int ret = 0;
 	struct iface *iface;
+	struct ifmatch state;
 
 	smclog(LOG_DEBUG, "Pruning %s from list of multicast routing interfaces", ifname);
-	iface = iface_find_by_name(ifname);
-	if (!iface)
-		return 1;
-
-	ret = mroute4_del_vif(iface);
+	iface_match_init(&state);
+	while ((iface = iface_match_by_name(ifname, &state))) {
+		ret += mroute4_del_vif(iface);
 #ifdef HAVE_IPV6_MULTICAST_ROUTING
-	ret += mroute6_del_mif(iface);
+		ret += mroute6_del_mif(iface);
 #endif
+	}
 
-	return ret;
+	if (!state.match_count)
+		return 1;
+	else
+		return ret;
 }
 
 #ifdef ENABLE_CLIENT
@@ -1124,12 +1210,12 @@ static int show_mroute(int sd, struct mroute4 *r, int detail)
 	snprintf(buf, sizeof(buf), "%-34s %-16s", sg, i->name);
 
 	if (detail) {
-		char stats[20];
+		char stats[30];
 		unsigned long p = 0, b = 0;
 
 		get_stats4(r, &p, &b, NULL);
-		snprintf(stats, sizeof(stats), " %7ld %8ld ", p, b);
-		strcat(buf, stats);
+		snprintf(stats, sizeof(stats), " %10lu %10lu ", p, b);
+		strlcat(buf, stats, sizeof(buf));
 	}
 
 	for (vif = 0; vif < MAX_MC_VIFS; vif++) {
@@ -1143,9 +1229,9 @@ static int show_mroute(int sd, struct mroute4 *r, int detail)
 			continue;
 
 		snprintf(tmp, sizeof(tmp), " %s", i->name);
-		strcat(buf, tmp);
+		strlcat(buf, tmp, sizeof(buf));
 	}
-	strcat(buf, "\n");
+	strlcat(buf, "\n", sizeof(buf));
 
 	if (ipc_send(sd, buf, strlen(buf)) < 0) {
 		smclog(LOG_ERR, "Failed sending reply to client: %s", strerror(errno));
